#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#define __rpcc __builtin_alpha_rpcc
#define __prefetch __builtin_prefetch

#define LENGTH 2*2 /* 2 unsigned longs, and 2 ints per long */

#ifdef __FIND_INLINE
#  define __kernel_cttz(x)      __builtin_ctzl(x)

static inline unsigned long __ffs(unsigned long word)
{
	/* Whee.  EV67 can calculate it directly.  */
	return __kernel_cttz(word);
}

unsigned long
sched_find_first_bit(const unsigned long b[2])
{
	unsigned long b0 = b[0], b1 = b[1]/*, b2 = b[2]*/;
	unsigned long ofs = 64;

	/*ofs = (b1 ? 64 : 128);
	b1 = (b1 ? b1 : b2);*/
	ofs = (b0 ? 0 : ofs);
	b0 = (b0 ? b0 : b1);

	return __ffs(b0) + ofs;
}
#else
extern unsigned long sched_find_first_bit(const unsigned long b[2]);
#endif

#ifdef __REWRITTEN_INLINE
unsigned long rewritten(const unsigned long b[2]) {
	unsigned long b0, b1;
	unsigned long ofs;
	unsigned long output, temp;

	asm volatile(
		"ldq	%1,0($16)	# b0\n"
		"clr	%3		# ofs\n"
		"ldq	%2,8($16)	# b1\n"
		"cmoveq	%1,64,%3	# ofs    = (b0 ? ofs : 64);\n"
		"cmoveq	%1,%2,%1	# output = (b0 ? b0 : b1);\n"
		"cttz	%1,%4		# temp = cttz(temp);\n"
		"addq	%4,%3,%0	# ret = temp + ofs\n"
		: "=&r" (output), "=&r" (b0), "=&r" (b1), "=&r" (ofs), "=&r" (temp)
	);
	return output;
}
#else
extern unsigned long rewritten(const unsigned long b[2]);
#endif

int main() {
	unsigned int array[LENGTH];
	
	unsigned long result_1;
	unsigned long result_2;
	
	unsigned int start_1, end_1;
	unsigned int start_2, end_2;

	unsigned long total_1 = 0, total_2 = 0;
	
	int i, j;
	
	srand(time(NULL));

	for (i = 0; i < 1000; i++) {
		array[0] = 0;
		array[1] = 0;
		array[2] = 0;
		array[3] = 0;
		for (j = 0; j < LENGTH; j++) {
			if (rand() % 2) { /* to get some more variation */
				array[j] = rand();
			}
		}
		__prefetch(array);
		
#ifdef __REWRITTEN_FIRST
		start_2 = __rpcc();
		result_2 = rewritten((const unsigned long *)array);
		end_2 = __rpcc();
#endif

		start_1 = __rpcc();
		result_1 = sched_find_first_bit((const unsigned long *)array);
		end_1 = __rpcc();

#ifndef __REWRITTEN_FIRST
		start_2 = __rpcc();
		result_2 = rewritten((const unsigned long *)array);
		end_2 = __rpcc();
#endif

		printf("input is 0x%08lx%08lx%08lx%08lx - result_1 is %lu, result_2 is %lu\n",
			array[0], array[1], array[2], array[3], result_1, result_2);
			
		assert(result_1 == result_2);
		total_1 += end_1 - start_1;
		total_2 += end_2 - start_2;
/*		printf("%04u - %04u\n", end_1 - start_1, end_2 - start_2);*/
	}
	printf("%lu - %lu\n", total_1, total_2);
#ifdef __REWRITTEN_INLINE
	puts("__REWRITTEN_INLINE");
#endif
#ifdef __REWRITTEN_FIRST
	puts("__REWRITTEN_FIRST");
#endif
#ifdef __FIND_INLINE
	puts("__FIND_INLINE");
#endif
	return 0;
}
